"""Interface for interacting with an agent."""
import asyncio
from contextlib import asynccontextmanager, suppress
import json
import logging
from typing import (
    List,
    NamedTuple,
    Optional,
    Tuple,
)

from acapy_client.api.connection import (
    create_invitation as _create_invitation,
    get_connection as _get_connection,
    get_connections as _get_connections,
    receive_invitation as _receive_invitation,
)
from acapy_client.api.credential_definition import publish_cred_def as _publish_cred_def
from acapy_client.api.issue_credential_v1_0 import (
    get_issue_credential_records as _get_cred_ex_records,
    get_issue_credential_records_cred_ex_id as _get_cred_ex_record,
)
from acapy_client.api.present_proof_v1_0 import (
    get_present_proof_records as _get_pres_ex_records,
    get_present_proof_records_pres_ex_id as _get_pres_ex_record,
)
from acapy_client.api.ledger import accept_taa as _accept_taa, fetch_taa as _fetch_taa
from acapy_client.api.revocation import (
    publish_revocations as _publish_revocations,
)
from acapy_client.api.schema import publish_schema as _publish_schema
from acapy_client.api.server import get_status_config
from acapy_client.api.wallet import (
    create_did as _create_did,
    get_wallet_did_public as _get_public_did,
    set_public_did as _set_public_did,
)
from acapy_client.client import Client
from acapy_client.models.connection_invitation import ConnectionInvitation
from acapy_client.models.connection_list import ConnectionList
from acapy_client.models.create_invitation_request import CreateInvitationRequest
from acapy_client.models.create_invitation_request_metadata import (
    CreateInvitationRequestMetadata,
)
from acapy_client.models.credential_definition_send_request import (
    CredentialDefinitionSendRequest,
)
from acapy_client.models.credential_definition_send_result import (
    CredentialDefinitionSendResult,
)
from acapy_client.models.did import DID
from acapy_client.models.did_create import DIDCreate
from acapy_client.models.did_result import DIDResult
from acapy_client.models.publish_revocations import PublishRevocations
from acapy_client.models.receive_invitation_request import ReceiveInvitationRequest
from acapy_client.models.schema_send_request import SchemaSendRequest
from acapy_client.models.schema_send_result import SchemaSendResult
from acapy_client.models.taa_accept import TAAAccept
from acapy_client.models.taa_info import TAAInfo
from acapy_client.models.taa_record import TAARecord
from acapy_client.models.taa_result import TAAResult
from acapy_client.models.txn_or_credential_definition_send_result import (
    TxnOrCredentialDefinitionSendResult,
)
from acapy_client.models.txn_or_schema_send_result import TxnOrSchemaSendResult
from acapy_client.types import Response, UNSET, Unset
from aiohttp import ClientSession
import aiohttp
from aiohttp.http_websocket import WSMsgType
from httpx import AsyncClient

from .api import Api
from .onboarding import get_onboarder
from .queue import Queue
from .connection import Connection
from .credential_exchange import CredentialExchange
from .presentation_exchange import PresentationExchange
from .utils import unwrap, unwrap_or


LOGGER = logging.getLogger(__name__)


class ControllerError(Exception):
    """Raised on error in controller."""


class Event(NamedTuple):
    topic: str
    payload: dict

    def __str__(self) -> str:
        payload = json.dumps(self.payload, sort_keys=True, indent=2)
        if len(payload) > 1000:
            payload = json.dumps(self.payload, sort_keys=True)
        return f'Event(topic="{self.topic}", payload={payload})'


class Controller:
    """Interface for interacting with an agent."""

    def __init__(self, name: str, base_url: str, *, headers: Optional[dict] = None):
        self.name = name
        self.client = Client(
            base_url=base_url, headers=headers or {}, timeout=5.0, verify_ssl=True
        )

        self.event_queue: Optional[Queue[Event]] = None
        self._ws_task: Optional[asyncio.Task] = None
        self._ws: Optional[aiohttp.ClientWebSocketResponse] = None

    @asynccontextmanager
    async def listening(self):
        if not self.event_queue:
            self.event_queue = Queue()

        self.event_queue.flush()
        self._ws_task = asyncio.get_event_loop().create_task(self.ws())
        yield

        if self._ws and not self._ws.closed:
            await self._ws.close()
        self._ws = None

        self._ws_task.cancel()
        with suppress(asyncio.CancelledError):
            await self._ws_task
        self._ws_task = None

    async def ws(self):
        assert self.event_queue
        async with ClientSession(self.client.base_url) as session:
            async with session.ws_connect("/ws", timeout=30.0) as ws:
                self._ws = ws
                async for msg in ws:
                    if msg.type == WSMsgType.TEXT:
                        data = msg.json()
                        if data.get("topic") != "ping":
                            try:
                                event = Event(**data)
                                LOGGER.debug("%s: Event received: %s", self.name, event)
                                await self.event_queue.put(event)
                            except Exception:
                                LOGGER.warning(
                                    "Unable to parse event: %s",
                                    json.dumps(data, indent=2),
                                )
                        elif data.get("topic") == "ping":
                            LOGGER.debug("%s: WS Ping received", self.name)
                    if msg.type == WSMsgType.ERROR:
                        break

    def clear_events(self):
        assert self.event_queue
        self.event_queue.flush()

    async def get_connections(self) -> ConnectionList:
        get_connections = Api(
            self.name, _get_connections._get_kwargs, _get_connections.asyncio_detailed
        )
        return await get_connections(client=self.client)

    async def get_connection(self, connection_id: str) -> Connection:
        get_connection = Api(
            self.name, _get_connection._get_kwargs, _get_connection.asyncio_detailed
        )
        record = await get_connection(client=self.client, conn_id=connection_id)
        return Connection(self, unwrap(record.connection_id), record)

    async def receive_invitation(
        self,
        invite: ConnectionInvitation,
        *,
        auto_accept: bool = False,
        alias: Optional[str] = None,
    ) -> Connection:
        receive_invitation = Api(
            self.name,
            _receive_invitation._get_kwargs,
            _receive_invitation.asyncio_detailed,
        )
        record = await receive_invitation(
            client=self.client,
            json_body=ReceiveInvitationRequest.from_dict(invite.to_dict()),
            auto_accept=auto_accept,
            alias=alias,
        )
        return Connection(self, unwrap(record.connection_id), record)

    async def create_invitation(
        self,
        *,
        auto_accept: bool = False,
        alias: Optional[str] = None,
        metadata: Optional[dict] = None,
    ) -> Tuple[Connection, ConnectionInvitation]:
        create_invitation = Api(
            self.name,
            _create_invitation._get_kwargs,
            _create_invitation.asyncio_detailed,
        )
        invitation_result = await create_invitation(
            client=self.client,
            json_body=CreateInvitationRequest(
                metadata=CreateInvitationRequestMetadata.from_dict(metadata or {})
            ),
            auto_accept=auto_accept,
            alias=alias,
        )
        return (
            Connection(
                self, unwrap(invitation_result.connection_id), invitation_result
            ),
            unwrap(invitation_result.invitation),
        )

    async def publish_schema(
        self, schema_name: str, schema_version: str, attributes: List[str]
    ) -> str:
        publish_schema = Api(
            self.name, _publish_schema._get_kwargs, _publish_schema.asyncio_detailed
        )
        result = await publish_schema(
            client=self.client,
            json_body=SchemaSendRequest(
                schema_name=schema_name,
                schema_version=schema_version,
                attributes=attributes,
            ),
        )
        if isinstance(result, SchemaSendResult):
            return result.schema_id

        assert isinstance(result, TxnOrSchemaSendResult)
        return unwrap(result.sent).schema_id

    async def publish_cred_def(
        self,
        schema_id: str,
        *,
        tag: Optional[str] = None,
        support_revocation: bool = False,
        revocation_registry_size: Optional[int] = None,
    ) -> str:
        """Publish cred def and return cred def id."""
        publish_cred_def = Api(
            self.name, _publish_cred_def._get_kwargs, _publish_cred_def.asyncio_detailed
        )
        result = await publish_cred_def(
            client=self.client.with_timeout(30.0),
            json_body=CredentialDefinitionSendRequest(
                schema_id=schema_id,
                revocation_registry_size=revocation_registry_size or UNSET,
                support_revocation=support_revocation,
                tag=tag or UNSET,
            ),
        )
        if isinstance(result, CredentialDefinitionSendResult):
            return result.credential_definition_id

        assert isinstance(result, TxnOrCredentialDefinitionSendResult)
        return unwrap(result.sent).credential_definition_id

    async def get_cred_ex_records(self) -> List[CredentialExchange]:
        get_cred_ex_records = Api(
            self.name,
            _get_cred_ex_records._get_kwargs,
            _get_cred_ex_records.asyncio_detailed,
        )
        result = await get_cred_ex_records(client=self.client)
        return [
            CredentialExchange(
                self,
                unwrap(record.connection_id),
                unwrap(record.credential_exchange_id),
                record,
            )
            for record in unwrap_or(result.results) or []
        ]

    async def get_cred_ex_record(self, cred_ex_id: str) -> CredentialExchange:
        get_cred_ex_record = Api(
            self.name,
            _get_cred_ex_record._get_kwargs,
            _get_cred_ex_record.asyncio_detailed,
        )
        result = await get_cred_ex_record(cred_ex_id, client=self.client)
        return CredentialExchange(
            self,
            unwrap(result.connection_id),
            unwrap(result.credential_exchange_id),
            result,
        )

    async def get_pres_ex_records(self) -> List[PresentationExchange]:
        get_pres_ex_records = Api(
            self.name,
            _get_pres_ex_records._get_kwargs,
            _get_pres_ex_records.asyncio_detailed,
        )
        result = await get_pres_ex_records(client=self.client)
        return [
            PresentationExchange(
                self,
                unwrap(record.connection_id),
                unwrap(record.presentation_exchange_id),
                record,
            )
            for record in unwrap_or(result.results) or []
        ]

    async def get_pres_ex_record(self, pres_ex_id: str) -> PresentationExchange:
        get_pres_ex_record = Api(
            self.name,
            _get_pres_ex_record._get_kwargs,
            _get_pres_ex_record.asyncio_detailed,
        )
        result = await get_pres_ex_record(pres_ex_id, client=self.client)
        return PresentationExchange(
            self,
            unwrap(result.connection_id),
            unwrap(result.presentation_exchange_id),
            result,
        )

    async def publish_revocations(self):
        publish_revocations = Api(
            self.name,
            _publish_revocations._get_kwargs,
            _publish_revocations.asyncio_detailed,
        )
        await publish_revocations(client=self.client, json_body=PublishRevocations())

    async def fetch_taa(self) -> TAAInfo:
        # TODO Fix
        async def _manual_override(*, client: Client):
            async with AsyncClient(base_url=self.client.base_url) as session:
                result = await session.get("/ledger/taa")
                info = result.json().get("result", {})
                return Response(
                    status_code=result.status_code,
                    content=result.content,
                    headers=result.headers,
                    parsed=TAAResult(
                        result=TAAInfo.from_dict(
                            {
                                "aml_record": info.get("aml_record") or UNSET,
                                "taa_record": info.get("taa_record") or UNSET,
                                "taa_accepted": info.get("taa_accepted"),
                                "taa_required": info.get("taa_required"),
                            }
                        )
                    ),
                )

        fetch_taa = Api(self.name, _fetch_taa._get_kwargs, _manual_override)
        return unwrap((await fetch_taa(client=self.client)).result)

    async def accept_taa(self, text: str, version: str):
        accept_taa = Api(
            self.name,
            _accept_taa._get_kwargs,
            _accept_taa.asyncio_detailed,
        )
        await accept_taa(
            client=self.client,
            json_body=TAAAccept(
                mechanism="on_file",
                text=text,
                version=version,
            ),
        )

    async def create_did(self) -> Tuple[str, str]:
        create_did = Api(
            self.name, _create_did._get_kwargs, _create_did.asyncio_detailed
        )
        result = unwrap(
            (await create_did(client=self.client, json_body=DIDCreate())).result
        )
        return unwrap(result.did), unwrap(result.verkey)

    async def get_public_did(self) -> Optional[str]:
        # TODO Fix
        async def _manual_override(*, client: Client):
            async with AsyncClient(base_url=self.client.base_url) as session:
                result = await session.get("/wallet/did/public")
                did = result.json().get("result")
                return Response(
                    status_code=result.status_code,
                    content=result.content,
                    headers=result.headers,
                    parsed=DIDResult(result=DID.from_dict(did) if did else UNSET),
                )

        get_public_did = Api(self.name, _get_public_did._get_kwargs, _manual_override)
        result = await get_public_did(client=self.client)
        if isinstance(result.result, Unset):
            return None

        return unwrap(result.result.did)

    async def get_genesis_url(self) -> Optional[str]:
        config = await get_status_config.asyncio(client=self.client)
        assert config
        config = unwrap(config.config).to_dict()
        return config.get("ledger.genesis_url")

    async def set_public_did(self, did: str):
        set_public_did = Api(
            self.name,
            _set_public_did._get_kwargs,
            _set_public_did.asyncio_detailed,
        )
        await set_public_did(client=self.client, did=did)

    async def onboard(self):
        genesis_url = await self.get_genesis_url()
        if not genesis_url:
            raise ControllerError("No ledger configured on agent")

        taa = await self.fetch_taa()
        if taa.taa_required is True and (
            isinstance(taa.taa_accepted, (Unset)) or taa.taa_accepted is None
        ):
            assert isinstance(taa.taa_record, TAARecord)
            await self.accept_taa(
                unwrap(taa.taa_record.text), unwrap(taa.taa_record.version)
            )

        public_did = await self.get_public_did()
        if not public_did:
            public_did, verkey = await self.create_did()
            onboarder = get_onboarder(genesis_url)
            if not onboarder:
                raise ControllerError(
                    "Unrecognized ledger, cannot automatically onboard"
                )
            await onboarder.onboard(public_did, verkey)
            await self.set_public_did(public_did)



"""Interface for interacting with an agent."""
import json
import logging
from typing import (
    TYPE_CHECKING,
    List,
)

from acapy_client.api.present_proof_v1_0 import (
    get_present_proof_records_pres_ex_id_credentials as _fetch_relevant_credentials,
    post_present_proof_records_pres_ex_id_problem_report as _pres_ex_problem_report,
    post_present_proof_records_pres_ex_id_send_presentation as _send_presentation,
    post_present_proof_records_pres_ex_id_verify_presentation as _verify_presentation,
)
from acapy_client.client import Client
from acapy_client.models.indy_cred_info import IndyCredInfo
from acapy_client.models.indy_cred_precis import IndyCredPrecis
from acapy_client.models.indy_non_revocation_interval import IndyNonRevocationInterval
from acapy_client.models.indy_pres_spec import IndyPresSpec
from acapy_client.models.v10_presentation_exchange import V10PresentationExchange
from acapy_client.models.v10_presentation_problem_report_request import (
    V10PresentationProblemReportRequest,
)
from acapy_client.types import Response, UNSET
from httpx import AsyncClient

from .api import Api
from .record import Record
from .utils import unwrap, unwrap_or

if TYPE_CHECKING:
    from .controller import Controller, Event


LOGGER = logging.getLogger(__name__)


class PresentationExchange(Record[V10PresentationExchange]):
    topic = "present_proof"

    def __init__(
        self,
        controller: "Controller",
        connection_id: str,
        presentation_exchange_id: str,
        record: V10PresentationExchange,
    ):
        super().__init__(controller, record)
        self.connection_id = connection_id
        self.presentation_exchange_id = presentation_exchange_id
        self.record = record

    @property
    def name(self) -> str:
        return f"{self.controller.name} Pres Ex ({self.presentation_exchange_id})"

    def _state_condition(self, event: "Event") -> bool:
        return (
            event.payload["connection_id"] == self.connection_id
            and event.payload["presentation_exchange_id"]
            == self.presentation_exchange_id
        )

    def summary(self) -> str:
        comment = unwrap_or(unwrap(self.record.presentation_request_dict).comment, None)
        return f"{self.name} Summary: " + json.dumps(
            {
                "state": self.record.state,
                "verified": unwrap_or(self.record.verified, None),
                "presentation_request": unwrap(
                    self.record.presentation_request
                ).to_dict(),
                "comment": comment,
            },
            sort_keys=True,
            indent=2,
        )

    async def request_received(self):
        """Wait for record to advance to request_received state."""
        return await self.wait_for_state("request_received")

    async def presentation_received(self):
        """Wait for record to advance to presentation_received state."""
        return await self.wait_for_state("presentation_received")

    async def verified(self):
        """Wait for record to advance to verified state."""
        return await self.wait_for_state("verified")

    async def presentation_acked(self):
        """Wait for record to advance to presentation_acked state."""
        return await self.wait_for_state("presentation_acked")

    async def fetch_relevant_credentials(self) -> List[IndyCredPrecis]:
        """Fetch credentials that could be used to fulfill this presentation request."""
        # TODO fix
        async def _manual_override(presentation_exchange_id: str, *, client: Client):
            async with AsyncClient(base_url=self.client.base_url) as session:
                result = await session.get(
                    f"/present-proof/records/{presentation_exchange_id}/credentials"
                )
                creds = result.json()
                return Response(
                    status_code=result.status_code,
                    content=result.content,
                    parsed=[
                        IndyCredPrecis(
                            cred_info=IndyCredInfo.from_dict(value["cred_info"]),
                            interval=IndyNonRevocationInterval.from_dict(
                                value["interval"]
                            )
                            if value.get("interval")
                            else UNSET,
                            presentation_referents=value.get("presentation_referents")
                            or [],
                        )
                        for value in creds
                    ],
                    headers=result.headers,
                )

        fetch_relevant_credentials = Api(
            self.name,
            _fetch_relevant_credentials._get_kwargs,
            _manual_override,
        )
        result = await fetch_relevant_credentials(
            self.presentation_exchange_id,
            client=self.client,
        )
        return result

    async def _problem_report(self, description: str):
        problem_report = Api(
            self.name,
            _pres_ex_problem_report._get_kwargs,
            _pres_ex_problem_report.asyncio_detailed,
        )
        await problem_report(
            self.presentation_exchange_id,
            client=self.client,
            json_body=V10PresentationProblemReportRequest(description=description),
        )

    async def reject(self, reason: str):
        return await self._problem_report(reason)

    async def abandon(self, reason: str):
        return await self._problem_report(reason)

    async def auto_prepare_presentation(
        self, fetched_credentials: List[IndyCredPrecis]
    ) -> IndyPresSpec:
        """Automatically fulfill a presentation.

        Uses the first available cred for each attribute and predicate.
        """
        request = unwrap(self.record.presentation_request)
        requested_attributes = {}
        for pres_referrent in request.requested_attributes.to_dict().keys():
            for cred_precis in fetched_credentials:
                if pres_referrent in unwrap(cred_precis.presentation_referents):
                    requested_attributes[pres_referrent] = {
                        "cred_id": unwrap(unwrap(cred_precis.cred_info).referent),
                        "revealed": True,
                    }
        requested_predicates = {}
        for pres_referrent in request.requested_predicates.to_dict().keys():
            for cred_precis in fetched_credentials:
                if pres_referrent in unwrap(cred_precis.presentation_referents):
                    requested_predicates[pres_referrent] = {
                        "cred_id": unwrap(unwrap(cred_precis.cred_info).referent),
                    }

        return IndyPresSpec.from_dict(
            {
                "requested_attributes": requested_attributes,
                "requested_predicates": requested_predicates,
                "self_attested_attributes": {},
            }
        )

    async def send_presentation(
        self, pres_spec: IndyPresSpec
    ) -> V10PresentationExchange:
        send_presentation = Api(
            self.name,
            _send_presentation._get_kwargs,
            _send_presentation.asyncio_detailed,
        )
        result = await send_presentation(
            self.presentation_exchange_id, client=self.client, json_body=pres_spec
        )
        self.record = result
        return result

    async def verify_presentation(self) -> V10PresentationExchange:
        verify_presentation = Api(
            self.name,
            _verify_presentation._get_kwargs,
            _verify_presentation.asyncio_detailed,
        )
        result = await verify_presentation(
            self.presentation_exchange_id, client=self.client
        )
        self.record = result
        return result



"""Definitions of protocol flows."""
from typing import Any, Dict, List, Optional, Tuple

from .connection import Connection
from .controller import Controller
from .credential_exchange import CredentialExchange
from .presentation_exchange import PresentationExchange


Pair = Tuple[Controller, Controller]
ConnectedPair = Tuple[Connection, Connection]
CredExPair = Tuple[CredentialExchange, CredentialExchange]
PresExPair = Tuple[PresentationExchange, PresentationExchange]


async def connect(pair: Pair):
    lhs, rhs = pair
    async with lhs.listening(), rhs.listening():
        lhs_conn, invite = await lhs.create_invitation()
        lhs.clear_events()

        rhs_conn = await rhs.receive_invitation(invite, auto_accept=False)

        await rhs_conn.accept_invitation()
        rhs.clear_events()

        await lhs_conn.request_received()
        await lhs_conn.accept_request()

        await rhs_conn.response_received()
        await rhs_conn.send_trust_ping()

        await lhs_conn.active()
        await rhs_conn.active()

        return lhs_conn, rhs_conn


async def issue_credential(pair: ConnectedPair, *, cred_def_id: str, **attributes):
    issuer, holder = pair
    async with issuer.listening(), holder.listening():
        issuer_cred_ex = await issuer.issue_credential(cred_def_id, **attributes)
        holder_cred_ex = await holder.receive_cred_ex()
        assert holder_cred_ex.record.state == "offer_received"
        await holder_cred_ex.send_request()
        await issuer_cred_ex.request_received()
        # Issuer automatically issues
        await holder_cred_ex.credential_received()
        await holder_cred_ex.store()
        await issuer_cred_ex.credential_acked()
        await holder_cred_ex.credential_acked()
        return issuer_cred_ex, holder_cred_ex


async def revoke_credential(
    pair: CredExPair, *, comment: Optional[str] = None, publish: bool = False
):
    issuer_cred_ex, holder_cred_ex = pair
    async with issuer_cred_ex.listening(), holder_cred_ex.listening():
        await issuer_cred_ex.revoke(comment=comment, publish=False)
        if publish:
            await issuer_cred_ex.controller.publish_revocations()
            await holder_cred_ex.receive_revocation_notification()
        return issuer_cred_ex, holder_cred_ex


async def present_proof(
    pair: ConnectedPair,
    *,
    name: Optional[str] = None,
    version: Optional[str] = None,
    comment: Optional[str] = None,
    requested_attributes: Optional[List[Dict[str, Any]]] = None,
    requested_predicates: Optional[List[Dict[str, Any]]] = None,
    non_revoked: Optional[Dict[str, int]] = None,
):
    verifier, holder = pair
    async with verifier.listening(), holder.listening():
        verifier_pres = await verifier.request_presentation(
            name=name,
            version=version,
            comment=comment,
            requested_attributes=requested_attributes,
            requested_predicates=requested_predicates,
            non_revoked=non_revoked,
        )
        holder_pres = await holder.receive_pres_ex()
        relevant_creds = await holder_pres.fetch_relevant_credentials()
        pres_spec = await holder_pres.auto_prepare_presentation(relevant_creds)
        await holder_pres.send_presentation(pres_spec)

        await verifier_pres.presentation_received()
        await verifier_pres.verify_presentation()
        await verifier_pres.verified()
        await holder_pres.presentation_acked()
        return verifier_pres, holder_pres



"""Interface for interacting with an agent."""
import asyncio
from dataclasses import dataclass
import json
import logging
from typing import TYPE_CHECKING, Optional

from acapy_client.api.issue_credential_v1_0 import (
    post_issue_credential_records_cred_ex_id_problem_report as _cred_ex_problem_report,
    post_issue_credential_records_cred_ex_id_send_request as _issue_cred_send_request,
    post_issue_credential_records_cred_ex_id_store as _issue_cred_store,
)
from acapy_client.api.revocation import revoke_credential as _revoke_credential
from acapy_client.models.revoke_request import RevokeRequest
from acapy_client.models.v10_credential_exchange import V10CredentialExchange
from acapy_client.models.v10_credential_problem_report_request import (
    V10CredentialProblemReportRequest,
)
from acapy_client.models.v10_credential_store_request import V10CredentialStoreRequest
from acapy_client.types import UNSET

from .api import Api
from .record import Record
from .utils import unwrap, unwrap_or

if TYPE_CHECKING:
    from .controller import Controller, Event


LOGGER = logging.getLogger(__name__)


@dataclass
class RevocationNotification:
    pass


class CredentialExchange(Record[V10CredentialExchange]):

    topic = "issue_credential"

    def __init__(
        self,
        controller: "Controller",
        connection_id: str,
        credential_exchange_id: str,
        record: V10CredentialExchange,
    ):
        super().__init__(controller, record)
        self.connection_id = connection_id
        self.credential_exchange_id = credential_exchange_id

    @property
    def name(self) -> str:
        return f"{self.controller.name} Cred Ex ({self.credential_exchange_id})"

    def summary(self) -> str:
        return f"{self.name} Summary: " + json.dumps(
            {
                "connection_id": self.connection_id,
                "credential_exchange_id": self.credential_exchange_id,
                "state": self.record.state,
                "credential_definition_id": self.record.credential_definition_id,
                "attributes": {
                    name: value["raw"]
                    for name, value in unwrap(self.record.raw_credential)
                    .to_dict()["values"]
                    .items()
                }
                if unwrap_or(self.record.raw_credential)
                else {},
            },
            indent=2,
            sort_keys=True,
        )

    def _state_condition(self, event: "Event") -> bool:
        return (
            event.payload["connection_id"] == self.connection_id
            and event.payload["credential_exchange_id"] == self.credential_exchange_id
        )

    async def _problem_report(self, description: str):
        problem_report = Api(
            self.name,
            _cred_ex_problem_report._get_kwargs,
            _cred_ex_problem_report.asyncio_detailed,
        )
        await problem_report(
            self.credential_exchange_id,
            client=self.client,
            json_body=V10CredentialProblemReportRequest(description=description),
        )

    async def reject(self, reason: str):
        return await self._problem_report(reason)

    async def abandon(self, reason: str):
        return await self._problem_report(reason)

    async def send_request(self):
        send_request = Api(
            self.name,
            _issue_cred_send_request._get_kwargs,
            _issue_cred_send_request.asyncio_detailed,
        )
        result = await send_request(self.credential_exchange_id, client=self.client)
        self.record = result

    async def store(self):
        store = Api(
            self.name, _issue_cred_store._get_kwargs, _issue_cred_store.asyncio_detailed
        )
        result = await store(
            self.credential_exchange_id,
            client=self.client,
            json_body=V10CredentialStoreRequest(),
        )
        self.record = result

    async def offer_received(self):
        return await self.wait_for_state("offer_received")

    async def request_received(self):
        return await self.wait_for_state("request_received")

    async def credential_received(self):
        return await self.wait_for_state("credential_received")

    async def credential_acked(self):
        return await self.wait_for_state("credential_acked")

    async def revoke(self, *, publish: bool = False, comment: Optional[str] = None):
        revoke_credential = Api(
            self.name,
            _revoke_credential._get_kwargs,
            _revoke_credential.asyncio_detailed,
        )
        await revoke_credential(
            client=self.client,
            json_body=RevokeRequest(
                comment=comment or UNSET,
                connection_id=self.connection_id,
                cred_ex_id=self.credential_exchange_id,
                publish=publish,
            ),
        )

    async def receive_revocation_notification(self) -> RevocationNotification:
        assert self.controller.event_queue
        LOGGER.info("%s awaiting notification of revocation...", self.name)
        try:
            event = await self.controller.event_queue.get(
                lambda event: event.topic == "revocation-notification"
            )
        except asyncio.TimeoutError:
            LOGGER.error(
                "Waiting for revocation notification timed out! "
                "Is --monitor-revocation-notification enabled on the holder? "
                "Is --notify-revocation enabled on the issuer?"
            )
            raise

        LOGGER.debug("%s: received event: %s", self.name, event)
        return RevocationNotification()



import asyncio
from typing import Any, Callable, Generic, List, Optional, Sequence, TypeVar


QueueEntry = TypeVar("QueueEntry")


class Queue(Generic[QueueEntry]):
    def __init__(
        self,
        *,
        condition: Optional[Callable[[QueueEntry], bool]] = None,
    ):
        self._queue: List[Any] = []
        self._cond: asyncio.Condition = asyncio.Condition()
        self.condition = condition

    def _first_matching_index(self, condition: Callable[[QueueEntry], bool]):
        for index, entry in enumerate(self._queue):
            if condition(entry):
                return index
        return None

    async def _get(
        self, condition: Optional[Callable[[QueueEntry], bool]] = None
    ) -> QueueEntry:
        """Retrieve a message from the queue."""
        while True:
            async with self._cond:
                # Lock acquired
                if not self._queue:
                    # No items on queue yet so we need to wait for items to show up
                    await self._cond.wait()

                if not self._queue:
                    # Another task grabbed the value before we got to it
                    continue

                if not condition:
                    # Just get the first message
                    return self._queue.pop()

                # Return first matching item, if present
                match_idx = self._first_matching_index(condition)
                if match_idx is not None:
                    return self._queue.pop(match_idx)
                else:
                    # Queue is not empty but no matching elements
                    # We need to wait for more before checking again
                    # Otherwise, this becomes a busy loop
                    await self._cond.wait()

    async def get(
        self,
        condition: Optional[Callable[[QueueEntry], bool]] = None,
        *,
        timeout: int = 5,
    ) -> QueueEntry:
        """Retrieve a message from the queue."""
        try:
            return await asyncio.wait_for(self._get(condition), timeout)
        except asyncio.TimeoutError:
            raise asyncio.TimeoutError("Retrieval from queue timed out") from None

    def get_all(
        self, condition: Optional[Callable[[QueueEntry], bool]] = None
    ) -> Sequence[QueueEntry]:
        """Return all messages matching a given condition."""
        messages = []
        if not self._queue:
            return messages

        if not condition:
            messages = [entry for entry in self._queue]
            self._queue.clear()
            return messages

        # Store messages that didn't match in the order they are seen
        filtered: List[QueueEntry] = []
        for entry in self._queue:
            if condition(entry):
                messages.append(entry)
            else:
                filtered.append(entry)

        # Queue contents set to messages that didn't match condition
        self._queue[:] = filtered
        return messages

    def get_nowait(
        self, condition: Optional[Callable[[QueueEntry], bool]] = None
    ) -> Optional[QueueEntry]:
        """Return a message from the queue without waiting."""
        if not self._queue:
            return None

        if not condition:
            return self._queue.pop()

        match_idx = self._first_matching_index(condition)
        if match_idx is not None:
            return self._queue.pop(match_idx)

        return None

    async def put(self, value: QueueEntry):
        """Push a message onto the queue and notify waiting tasks."""
        if not self.condition or self.condition(value):
            async with self._cond:
                self._queue.append(value)
                self._cond.notify_all()

    def flush(self) -> Sequence[QueueEntry]:
        """Clear queue and return final contents of queue at time of clear."""
        final = self._queue.copy()
        self._queue.clear()
        return final



from abc import ABC, abstractmethod
import json
from typing import Optional
from aiohttp import ClientSession

INDICIO_TESTNET_GENESIS = (
    "https://raw.githubusercontent.com/Indicio-tech/indicio-network/main/"
    "genesis_files/pool_transactions_testnet_genesis"
)
INDICIO_TESTNET_GENESIS_OLD = (
    "https://raw.githubusercontent.com/Indicio-tech/indicio-network/master/"
    "genesis_files/pool_transactions_testnet_genesis"
)


def get_onboarder(genesis_url: str) -> Optional["Onboarder"]:
    if genesis_url.endswith("/genesis"):
        # infer VonOnboarder
        return VonOnboarder(genesis_url.replace("/genesis", "register"))

    return {
        INDICIO_TESTNET_GENESIS: SelfServeOnboarder(
            "https://selfserve.indiciotech.io/nym", "testnet"
        ),
        INDICIO_TESTNET_GENESIS_OLD: SelfServeOnboarder(
            "https://selfserve.indiciotech.io/nym", "testnet"
        ),
    }.get(genesis_url)


class Onboarder(ABC):
    @abstractmethod
    async def onboard(self, did: str, verkey: str):
        """Onboard a DID"""


class VonOnboarder(Onboarder):
    def __init__(self, registration_url: str):
        self.registration_url = registration_url

    async def onboard(self, did: str, verkey: str):
        async with ClientSession() as session:
            async with session.post(
                self.registration_url,
                json={
                    "did": did,
                    "verkey": verkey,
                    "alias": None,
                    "role": "ENDORSER",
                },
            ) as resp:
                if resp.ok:
                    return await resp.json()


class SelfServeOnboarder(Onboarder):
    def __init__(self, registration_url: str, network: str):
        self.registration_url = registration_url
        self.network = network

    async def onboard(self, did: str, verkey: str):
        async with ClientSession() as session:
            async with session.post(
                self.registration_url,
                headers={"content-type": "application/json; charset=utf-8"},
                json={
                    "network": self.network,
                    "did": did,
                    "verkey": verkey,
                    "alias": None,
                    "role": "ENDORSER",
                },
            ) as resp:
                if resp.ok:
                    body = await resp.text()
                    try:
                        return json.loads(body)
                    except json.decoder.JSONDecodeError:
                        return None



import logging
from os import getenv
import sys


from blessings import Terminal

from .controller import flows
from .controller.connection import Connection
from .controller.controller import Controller
from .controller.credential_exchange import CredentialExchange
from .controller.presentation_exchange import PresentationExchange


LOG_LEVEL = getenv("LOG_LEVEL", "debug")


class ColorFormatter(logging.Formatter):
    def __init__(self, fmt: str):
        self.default = logging.Formatter(fmt)
        term = Terminal()
        self.formats = {
            logging.DEBUG: logging.Formatter(f"{term.dim}{fmt}{term.normal}"),
            logging.ERROR: logging.Formatter(f"{term.red}{fmt}{term.normal}"),
        }

    def format(self, record):
        formatter = self.formats.get(record.levelno, self.default)
        return formatter.format(record)


def logging_to_stdout():
    if sys.stdout.isatty():
        logger = logging.getLogger("acapy_revocation_demo")
        logger.setLevel(LOG_LEVEL.upper())
        ch = logging.StreamHandler()
        ch.setLevel(LOG_LEVEL.upper())
        ch.setFormatter(ColorFormatter("%(message)s"))
        logger.addHandler(ch)
    else:
        logging.basicConfig(
            stream=sys.stdout, level=logging.WARNING, format="%(message)s"
        )
        logging.getLogger("acapy_revocation_demo").setLevel(LOG_LEVEL.upper())


__all__ = [
    "Controller",
    "Connection",
    "PresentationExchange",
    "CredentialExchange",
    "flows",
]



"""A set of scenarios to be executed by the demo scripts."""

from os import getenv
import random
import string
import asyncio
import time
from typing import NamedTuple, Optional

from . import Controller, Connection, logging_to_stdout, flows

ISSUER = getenv("ISSUER", "http://host.docker.internal:8021")
VERIFIER = getenv("VERIFIER", "http://host.docker.internal:8031")
HOLDER = getenv("HOLDER", "http://host.docker.internal:8041")


def random_string(size):
    """Generate a random string."""
    return "".join(
        random.choice(string.ascii_letters + string.digits) for _ in range(size)
    )


class IssuerHolder(NamedTuple):
    issuer: Connection
    holder: Connection


class VerifierHolder(NamedTuple):
    verifier: Connection
    holder: Connection


async def connected(lhs: Controller, rhs: Controller):
    """Connect two agents."""
    return await flows.connect((lhs, rhs))


async def connected_issuer_holder(
    issuer_url: Optional[str] = None, holder_url: Optional[str] = None
):
    """Connect issuer and holder."""
    issuer = Controller("issuer", issuer_url or ISSUER)
    holder = Controller("holder", holder_url or HOLDER)
    return await connected(issuer, holder)


async def connected_verifier_holder(
    verifier_url: Optional[str] = None, holder_url: Optional[str] = None
):
    """Connect verifier and holder."""
    verifier = Controller("verifier", verifier_url or VERIFIER)
    holder = Controller("holder", holder_url or HOLDER)
    return await connected(verifier, holder)


async def prepare_ledger_artifacts(issuer: Controller, *, revocable: bool = False):
    """Prepare ledger artifacts for issuing a credential."""
    await issuer.onboard()
    schema_id = await issuer.publish_schema(
        random_string(8), "0.1", ["attr0", "attr1", "attr2"]
    )
    cred_def_id = await issuer.publish_cred_def(schema_id, support_revocation=revocable)
    return schema_id, cred_def_id


async def issued_credential(
    issuer: Connection, holder: Connection, *, revocable: bool = False
):
    """Issue credential to holder."""
    _, cred_def_id = await prepare_ledger_artifacts(
        issuer.controller, revocable=revocable
    )
    return await flows.issue_credential(
        (issuer, holder),
        cred_def_id=cred_def_id,
        attr0="test0",
        attr1="test1",
        attr2="test2",
    )


async def revoked_credential(issuer: Connection, holder: Connection):
    issuer_cred_ex, holder_cred_ex = await issued_credential(
        issuer, holder, revocable=True
    )
    return await flows.revoke_credential(
        (issuer_cred_ex, holder_cred_ex), comment="revoked by demo script", publish=True
    )


async def presented_proof(
    issuer_holder: flows.ConnectedPair, verifier_holder: flows.ConnectedPair
):
    """Proof presented to verifier from holder."""
    await issued_credential(*issuer_holder)
    pres_exes = await flows.present_proof(
        verifier_holder, requested_attributes=[{"name": "attr0"}]
    )
    for pres_ex in pres_exes:
        print(pres_ex.summary())

    return pres_exes


async def present_revoked_credential(
    issuer_holder: flows.ConnectedPair, verifier_holder: flows.ConnectedPair
):
    """Present a credential that has been revoked."""
    issuer_cred_ex, _ = await revoked_credential(*issuer_holder)
    now = int(time.time())
    pres_exes = await flows.present_proof(
        verifier_holder,
        comment="presentation after revocation",
        requested_attributes=[
            {
                "name": "attr0",
                "restrictions": [
                    {"cred_def_id": issuer_cred_ex.record.credential_definition_id}
                ],
            }
        ],
        non_revoked={"from": now, "to": now},
    )
    for pres_ex in pres_exes:
        print(pres_ex.summary())

    return pres_exes


async def main():
    logging_to_stdout()

    issuer_holder = await connected_issuer_holder()
    verifier_holder = await connected_verifier_holder()
    await present_revoked_credential(issuer_holder, verifier_holder)


if __name__ == "__main__":
    asyncio.run(main())



"""Interface for interacting with an agent."""
from abc import abstractproperty
import json
import logging
from typing import (
    Callable,
    ClassVar,
    Generic,
    Optional,
    Protocol,
    TYPE_CHECKING,
    Type,
    TypeVar,
)

from acapy_client.client import Client

if TYPE_CHECKING:
    from .controller import Controller, Event

LOGGER = logging.getLogger(__name__)


T = TypeVar("T")


class RecordProtocol(Protocol):
    @classmethod
    def from_dict(cls: Type[T], src_dict: dict) -> T:
        ...


RecordType = TypeVar("RecordType", bound=RecordProtocol)


class Record(Generic[RecordType]):
    """Base class for record like objects."""

    topic: ClassVar[str]

    def __init__(self, controller: "Controller", record: RecordType):
        self.controller = controller
        self.record = record

    @property
    def client(self) -> Client:
        return self.controller.client

    @abstractproperty
    def name(self) -> str:
        """Return name of object."""
        ...

    def _state_condition(self, event: "Event") -> bool:
        """Condition for matching states to this record."""
        return True

    def listening(self):
        return self.controller.listening()

    async def wait_for_state(
        self,
        state: str,
        *,
        condition: Optional[Callable[["Event"], bool]] = None,
        state_attribute: Optional[str] = None,
    ) -> "Event":
        """Wait for a given record state."""
        assert self.controller.event_queue
        LOGGER.info(
            "%s: %s awaiting state %s...", self.name, type(self).__name__, state
        )
        event = await self.controller.event_queue.get(
            lambda event: event.topic == self.topic
            and event.payload[state_attribute or "state"] == state
            and self._state_condition(event)
            and (condition(event) if condition else True),
            timeout=3,
        )
        LOGGER.info("%s: %s reached state %s", self.name, type(self).__name__, state)
        LOGGER.debug(
            "%s record state: %s",
            type(self).__name__,
            json.dumps(event.payload, sort_keys=True),
        )
        self.record = type(self.record).from_dict(event.payload)
        return event



"""Run the demo."""

import asyncio
from contextlib import contextmanager
import os
from os import getenv
import sys
import time

from blessings import Terminal

from . import Controller, logging_to_stdout
from .scenarios import connected


ISSUER = getenv("ISSUER", "http://localhost:3003")
VERIFIER = getenv("VERIFIER", "http://localhost:3005")
HOLDER = getenv("HOLDER", "http://localhost:3001")
LOG_LEVEL = getenv("LOG_LEVEL", "debug")

term = Terminal()


@contextmanager
def section(title: str):
    if sys.stdout.isatty():
        size = os.get_terminal_size()
        left = "=" * (int(size.columns / 2) - int((len(title) + 1) / 2))
        right = "=" * (size.columns - (len(left) + len(title) + 2))
        print(f"{term.blue}{term.bold}{left} {title} {right}{term.normal}")
    else:
        print(title)
    yield


async def main():
    """Run steps."""
    logging_to_stdout()

    issuer = Controller("issuer", ISSUER)
    holder = Controller("holder", HOLDER)

    with section("Establish Connection"):
        issuer_conn, holder_conn = await connected(issuer, holder)

    with section("Prepare for writing to the ledger"):
        await issuer.onboard()

    with section("Prepare credential ledger artifacts"):
        schema_id = await issuer.publish_schema(
            attributes=["firstname", "age"],
            schema_name="revocation_testing",
            schema_version="0.1.0",
        )
        cred_def_id = await issuer.publish_cred_def(schema_id, support_revocation=True)

    with section("Issue credential and request presentation"):
        async with issuer.listening(), holder.listening():
            issuer_cred_ex = await issuer_conn.issue_credential(
                cred_def_id=cred_def_id,
                firstname="Bob",
                age="42",
            )
            holder_cred_ex = await holder_conn.receive_cred_ex()
            await holder_cred_ex.send_request()
            await issuer_cred_ex.request_received()
            await holder_cred_ex.credential_received()
            await holder_cred_ex.store()
            await issuer_cred_ex.credential_acked()
            await holder_cred_ex.credential_acked()

        non_revoked_time = int(time.time())
        async with issuer.listening(), holder.listening():
            issuer_pres = await issuer_conn.request_presentation(
                comment="Before revocation (should verify true)",
                requested_attributes=[
                    {
                        "name": "firstname",
                        "restrictions": [{"cred_def_id": cred_def_id}],
                    }
                ],
                non_revoked={"from": non_revoked_time, "to": non_revoked_time},
            )
            holder_pres = await holder_conn.receive_pres_ex()
            relevant_creds = await holder_pres.fetch_relevant_credentials()
            pres_spec = await holder_pres.auto_prepare_presentation(relevant_creds)
            await holder_pres.send_presentation(pres_spec)

            await issuer_pres.presentation_received()
            await issuer_pres.verify_presentation()
            await issuer_pres.verified()
            await holder_pres.presentation_acked()

    with section("Revoke credential"):
        async with issuer.listening(), holder.listening():
            await issuer_cred_ex.revoke(publish=False, comment="revoked by demo")
            await issuer.publish_revocations()
            await holder_cred_ex.receive_revocation_notification()
            print("Waiting 10 seconds for revocation to propagate...")
            time.sleep(10)

    with section("Request proof from holder again after revoking"):
        before_revoking_time = non_revoked_time
        async with issuer.listening(), holder.listening():
            non_revoked_time = int(time.time())
            issuer_pres = await issuer_conn.request_presentation(
                comment="After revoking (should verify false)",
                requested_attributes=[
                    {
                        "name": "firstname",
                        "restrictions": [{"cred_def_id": cred_def_id}],
                    }
                ],
                non_revoked={"from": non_revoked_time - 1, "to": non_revoked_time},
            )
            holder_pres = await holder_conn.receive_pres_ex()
            relevant_creds = await holder_pres.fetch_relevant_credentials()
            pres_spec = await holder_pres.auto_prepare_presentation(relevant_creds)
            await holder_pres.send_presentation(pres_spec)

            await issuer_pres.presentation_received()
            await issuer_pres.verify_presentation()
            await issuer_pres.verified()
            await holder_pres.presentation_acked()

    with section(
        "Attempt another proof with non_revoked interval to before revocation"
    ):
        async with issuer.listening(), holder.listening():
            issuer_pres = await issuer_conn.request_presentation(
                comment="After revoking, interval before revocation (should verify true)",
                requested_attributes=[
                    {
                        "name": "firstname",
                        "restrictions": [{"cred_def_id": cred_def_id}],
                    }
                ],
                non_revoked={"from": before_revoking_time, "to": before_revoking_time},
            )
            holder_pres = await holder_conn.receive_pres_ex()
            relevant_creds = await holder_pres.fetch_relevant_credentials()
            pres_spec = await holder_pres.auto_prepare_presentation(relevant_creds)
            await holder_pres.send_presentation(pres_spec)

            await issuer_pres.presentation_received()
            await issuer_pres.verify_presentation()
            await issuer_pres.verified()
            await holder_pres.presentation_acked()

    with section("Attempt another proof with no non_revoked interval"):
        async with issuer.listening(), holder.listening():
            issuer_pres = await issuer_conn.request_presentation(
                comment="After revoking, no revocation interval (should verify true)",
                requested_attributes=[
                    {
                        "name": "firstname",
                        "restrictions": [{"cred_def_id": cred_def_id}],
                    }
                ],
            )
            holder_pres = await holder_conn.receive_pres_ex()
            relevant_creds = await holder_pres.fetch_relevant_credentials()
            pres_spec = await holder_pres.auto_prepare_presentation(relevant_creds)
            await holder_pres.send_presentation(pres_spec)

            await issuer_pres.presentation_received()
            await issuer_pres.verify_presentation()
            await issuer_pres.verified()
            await holder_pres.presentation_acked()

    with section(
        "Attempt another proof with non_revoked interval and local non_revoked override"
    ):
        non_revoked_time = int(time.time())
        async with issuer.listening(), holder.listening():
            issuer_pres = await issuer_conn.request_presentation(
                comment=(
                    "After revocation, non_revoked interval and local "
                    "non_revoked override (should verify true)"
                ),
                requested_attributes=[
                    {
                        "name": "firstname",
                        "restrictions": [{"cred_def_id": cred_def_id}],
                        "non_revoked": {
                            "from": before_revoking_time,
                            "to": before_revoking_time,
                        },
                    }
                ],
                non_revoked={"from": non_revoked_time, "to": non_revoked_time},
            )
            holder_pres = await holder_conn.receive_pres_ex()
            relevant_creds = await holder_pres.fetch_relevant_credentials()
            pres_spec = await holder_pres.auto_prepare_presentation(relevant_creds)
            await holder_pres.send_presentation(pres_spec)

            await issuer_pres.presentation_received()
            await issuer_pres.verify_presentation()
            await issuer_pres.verified()
            await holder_pres.presentation_acked()

    with section("Attempt another proof with only local non_revoked interval"):
        non_revoked_time = int(time.time())
        async with issuer.listening(), holder.listening():
            issuer_pres = await issuer_conn.request_presentation(
                comment=(
                    "After revocation, local non_revoked interval only "
                    "(should verify false)"
                ),
                requested_attributes=[
                    {
                        "name": "firstname",
                        "restrictions": [{"cred_def_id": cred_def_id}],
                        "non_revoked": {
                            "from": non_revoked_time,
                            "to": non_revoked_time,
                        },
                    }
                ],
            )
            holder_pres = await holder_conn.receive_pres_ex()
            relevant_creds = await holder_pres.fetch_relevant_credentials()
            pres_spec = await holder_pres.auto_prepare_presentation(relevant_creds)
            await holder_pres.send_presentation(pres_spec)

            await issuer_pres.presentation_received()
            await issuer_pres.verify_presentation()
            await issuer_pres.verified()
            await holder_pres.presentation_acked()

    with section("Query presentations"):
        presentations = await issuer_conn.get_pres_ex_records()

    with section("Presentation Summary"):
        for pres in presentations:
            print(pres.summary())


if __name__ == "__main__":
    asyncio.get_event_loop().run_until_complete(main())



"""Interface for interacting with an agent."""
import logging
from typing import (
    Optional,
    TypeVar,
    Union,
)

from acapy_client.types import Unset


LOGGER = logging.getLogger(__name__)


T = TypeVar("T")


def unwrap(value: Union[Unset, T]) -> T:
    """Unwrap a potentially unset value."""
    assert not isinstance(value, Unset)
    return value


def unwrap_or(value: Union[Unset, T], default: Optional[T] = None) -> Optional[T]:
    """Unwrap a potentially unset value, returning a default if given."""
    if not isinstance(value, Unset):
        return value
    return default



import json
import logging
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Generic,
    List,
    ParamSpec,
    Protocol,
    TypeVar,
    Union,
    cast,
)

from acapy_client.types import Response
from urllib.parse import urlparse


LOGGER = logging.getLogger(__name__)


class ApiError(Exception):
    """Raised when error on API call."""


class ResultProtocol(Protocol):
    def to_dict(self) -> dict:
        ...


ApiParams = ParamSpec("ApiParams")
ApiResult = TypeVar("ApiResult", bound=Union[List[ResultProtocol], ResultProtocol])


def _serialize(
    value: Union[List[ResultProtocol], ResultProtocol, None]
) -> Union[Dict, List, None]:
    """Serialize result."""
    if value is None:
        return value
    if isinstance(value, List):
        return [res.to_dict() for res in value]
    return value.to_dict()


class Api(Generic[ApiParams, ApiResult]):
    def __init__(
        self,
        name: str,
        request_builder: Callable[..., Dict[str, Any]],
        api: Callable[ApiParams, Coroutine[Any, Any, Response[ApiResult]]],
    ):
        self.name = name
        self.request_builder = request_builder
        self.api = api

    async def __call__(
        self, *args: ApiParams.args, **kwargs: ApiParams.kwargs
    ) -> ApiResult:
        request = self.request_builder(*args, **kwargs)
        path = urlparse(request["url"]).path
        method = cast(str, request["method"]).upper()
        body = request.get("json", {})
        LOGGER.info(
            "Request to %s %s %s: %s",
            self.name,
            method,
            path,
            json.dumps(body, sort_keys=True, indent=2),
        )
        result: Response = await self.api(*args, **kwargs)
        if result.status_code == 200:
            serialized = _serialize(result.parsed)
            response_out = json.dumps(serialized, indent=2, sort_keys=True)
            if response_out.count("\n") > 30:
                response_out = json.dumps(serialized, sort_keys=True)
            LOGGER.info("Response: %s", response_out)
        else:
            raise ApiError("Request failed!", result.status_code, result.content)

        assert result.parsed is not None
        return result.parsed



"""Interface for interacting with an agent."""
import json
import logging
from typing import Any, Dict, List, Optional, TYPE_CHECKING, TypeVar, Union

from acapy_client.api.connection import (
    get_metadata as _get_metadata,
    post_connections_conn_id_accept_invitation as _accept_invitation,
    post_connections_conn_id_accept_request as _accept_request,
    set_metadata as _set_metadata,
)
from acapy_client.api.issue_credential_v1_0 import (
    issue_credential_automated as _issue_credential,
    get_issue_credential_records as _get_cred_ex_records,
)
from acapy_client.api.present_proof_v1_0 import (
    send_proof_request as _send_proof_request,
    get_present_proof_records as _get_pres_ex_records,
)
from acapy_client.api.trustping import (
    post_connections_conn_id_send_ping as _send_trust_ping,
)
from acapy_client.models.conn_record import ConnRecord
from acapy_client.models.connection_metadata_set_request import (
    ConnectionMetadataSetRequest,
)
from acapy_client.models.cred_attr_spec import CredAttrSpec
from acapy_client.models.credential_preview import CredentialPreview
from acapy_client.models.indy_proof_request import IndyProofRequest
from acapy_client.models.indy_proof_request_non_revoked import (
    IndyProofRequestNonRevoked,
)
from acapy_client.models.indy_proof_request_requested_attributes import (
    IndyProofRequestRequestedAttributes,
)
from acapy_client.models.indy_proof_request_requested_predicates import (
    IndyProofRequestRequestedPredicates,
)
from acapy_client.models.invitation_result import InvitationResult
from acapy_client.models.ping_request import PingRequest
from acapy_client.models.v10_credential_exchange import V10CredentialExchange
from acapy_client.models.v10_credential_proposal_request_mand import (
    V10CredentialProposalRequestMand,
)
from acapy_client.models.v10_presentation_exchange import V10PresentationExchange
from acapy_client.models.v10_presentation_send_request_request import (
    V10PresentationSendRequestRequest,
)
from acapy_client.types import UNSET

from .api import Api
from .credential_exchange import CredentialExchange
from .presentation_exchange import PresentationExchange
from .record import Record
from .utils import unwrap, unwrap_or

if TYPE_CHECKING:
    from .controller import Controller


LOGGER = logging.getLogger(__name__)


T = TypeVar("T")


class Connection(Record[Union[InvitationResult, ConnRecord]]):
    topic = "connections"

    def __init__(
        self,
        controller: "Controller",
        connection_id: str,
        record: Union[InvitationResult, ConnRecord],
    ):
        super().__init__(controller, record)
        self.connection_id = connection_id

    @property
    def name(self) -> str:
        return f"{self.controller.name} ({self.connection_id})"

    async def wait_for_state(self, state: str):
        await super().wait_for_state(
            state=state,
            state_attribute="rfc23_state",
            condition=lambda event: event.payload["connection_id"]
            == self.connection_id,
        )

    async def invitation_received(self):
        await self.wait_for_state("invitation-received")

    async def request_received(self):
        await self.wait_for_state("request-received")

    async def response_received(self):
        await self.wait_for_state("response-received")

    async def completed(self):
        await self.wait_for_state("completed")

    async def active(self):
        await self.wait_for_state("completed")

    async def accept_invitation(self) -> ConnRecord:
        accept_invitation = Api(
            self.name,
            _accept_invitation._get_kwargs,
            _accept_invitation.asyncio_detailed,
        )
        result = await accept_invitation(client=self.client, conn_id=self.connection_id)
        self.record = result
        return result

    async def accept_request(self) -> ConnRecord:
        accept_request = Api(
            self.name, _accept_request._get_kwargs, _accept_request.asyncio_detailed
        )
        result = await accept_request(client=self.client, conn_id=self.connection_id)
        self.record = result
        return result

    async def get_metadata(self) -> dict:
        get_metadata = Api(
            self.name, _get_metadata._get_kwargs, _get_metadata.asyncio_detailed
        )
        return unwrap(
            (await get_metadata(client=self.client, conn_id=self.connection_id)).results
        ).to_dict()

    async def set_metadata(self, **metadata):
        set_metadata = Api(
            self.name, _set_metadata._get_kwargs, _set_metadata.asyncio_detailed
        )
        await set_metadata(
            self.connection_id,
            client=self.client,
            json_body=ConnectionMetadataSetRequest.from_dict({"metadata": metadata}),
        )

    async def send_trust_ping(self, comment: Optional[str] = None):
        send_trust_ping = Api(
            self.name, _send_trust_ping._get_kwargs, _send_trust_ping.asyncio_detailed
        )
        await send_trust_ping(
            self.connection_id,
            client=self.client,
            json_body=PingRequest(comment=comment),
        )

    async def issue_credential(
        self, cred_def_id: str, **attributes
    ) -> "CredentialExchange":
        issue_credential = Api(
            self.name, _issue_credential._get_kwargs, _issue_credential.asyncio_detailed
        )
        result = await issue_credential(
            client=self.client,
            json_body=V10CredentialProposalRequestMand(
                connection_id=self.connection_id,
                cred_def_id=cred_def_id,
                credential_proposal=CredentialPreview(
                    attributes=[
                        CredAttrSpec(name, value) for name, value in attributes.items()
                    ]
                ),
            ),
        )
        return CredentialExchange(
            self.controller,
            self.connection_id,
            unwrap(result.credential_exchange_id),
            result,
        )

    async def receive_cred_ex(self) -> "CredentialExchange":
        assert self.controller.event_queue
        LOGGER.info("%s: Connection awaiting credential exchange...", self.name)
        event = await self.controller.event_queue.get(
            lambda event: event.topic == CredentialExchange.topic
            and event.payload["connection_id"] == self.connection_id,
            timeout=3,
        )
        LOGGER.info(
            "CredentialExchange record from event: %s",
            json.dumps(event.payload, sort_keys=True, indent=2),
        )
        return CredentialExchange(
            self.controller,
            self.connection_id,
            credential_exchange_id=event.payload.get("credential_exchange_id"),
            record=V10CredentialExchange.from_dict(event.payload),
        )

    async def request_presentation(
        self,
        *,
        name: Optional[str] = None,
        version: Optional[str] = None,
        comment: Optional[str] = None,
        requested_attributes: Optional[List[Dict[str, Any]]] = None,
        requested_predicates: Optional[List[Dict[str, Any]]] = None,
        non_revoked: Optional[Dict[str, int]] = None,
    ) -> PresentationExchange:
        """Request a presentation from connection."""
        send_proof_request = Api(
            self.name,
            _send_proof_request._get_kwargs,
            _send_proof_request.asyncio_detailed,
        )
        result = await send_proof_request(
            client=self.client,
            json_body=V10PresentationSendRequestRequest(
                comment=comment or UNSET,
                connection_id=self.connection_id,
                proof_request=IndyProofRequest(
                    name=name or "proof",
                    version=version or "0.1.0",
                    requested_attributes=IndyProofRequestRequestedAttributes.from_dict(
                        {attr["name"]: attr for attr in requested_attributes or []}
                    ),
                    requested_predicates=IndyProofRequestRequestedPredicates.from_dict(
                        {pred["name"]: pred for pred in requested_predicates or []}
                    ),
                    non_revoked=IndyProofRequestNonRevoked.from_dict(non_revoked)
                    if non_revoked
                    else UNSET,
                ),
            ),
        )
        return PresentationExchange(
            self.controller,
            self.connection_id,
            unwrap(result.presentation_exchange_id),
            result,
        )

    async def receive_pres_ex(self) -> PresentationExchange:
        assert self.controller.event_queue
        LOGGER.info("%s: Connection awaiting presentation exchange...", self.name)
        event = await self.controller.event_queue.get(
            lambda event: event.topic == PresentationExchange.topic
            and event.payload["connection_id"] == self.connection_id,
            timeout=3,
        )
        LOGGER.info(
            "PresentationExchange record from event: %s",
            json.dumps(event.payload, sort_keys=True, indent=2),
        )
        return PresentationExchange(
            self.controller,
            self.connection_id,
            presentation_exchange_id=event.payload.get("presentation_exchange_id"),
            record=V10PresentationExchange.from_dict(event.payload),
        )

    async def get_cred_ex_records(self) -> List[CredentialExchange]:
        get_cred_ex_records = Api(
            self.name,
            _get_cred_ex_records._get_kwargs,
            _get_cred_ex_records.asyncio_detailed,
        )
        result = await get_cred_ex_records(
            client=self.client, connection_id=self.connection_id
        )
        return [
            CredentialExchange(
                self.controller,
                unwrap(record.connection_id),
                unwrap(record.credential_exchange_id),
                record,
            )
            for record in unwrap_or(result.results) or []
        ]

    async def get_pres_ex_records(self) -> List[PresentationExchange]:
        get_pres_ex_records = Api(
            self.name,
            _get_pres_ex_records._get_kwargs,
            _get_pres_ex_records.asyncio_detailed,
        )
        result = await get_pres_ex_records(
            client=self.client, connection_id=self.connection_id
        )
        return [
            PresentationExchange(
                self.controller,
                unwrap(record.connection_id),
                unwrap(record.presentation_exchange_id),
                record,
            )
            for record in unwrap_or(result.results) or []
        ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['acapy_revocation_demo', 'acapy_revocation_demo.controller']

package_data = \
{'': ['*']}

install_requires = \
['acapy-client==0.7.3.post0', 'aiohttp>=3.8.1,<4.0.0', 'blessings>=1.7,<2.0']

setup_kwargs = {
    'name': 'acapy-revocation-demo',
    'version': '0.1.0',
    'description': 'Demonstrate Indy AnonCreds revocation using ACA-Py',
    'long_description': None,
    'author': 'Daniel Bluhm',
    'author_email': 'dbluhm@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
